@@ -19,7 +19,6 @@ import (
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"math"
 	"net/http"
 	"net/http/httptest"
@@ -57,7 +56,7 @@ func TestNewScrapePool(t *testing.T) {
 	var (
 		app   = &nopAppendable{}
 		cfg   = &config.ScrapeConfig{}
-		sp, _ = newScrapePool(cfg, app, 0, nil, false)
+		sp, _ = newScrapePool(cfg, app, 0, nil, false, false, nil)
 	)
 
 	if a, ok := sp.appendable.(*nopAppendable); !ok || a != app {
@@ -92,7 +91,7 @@ func TestDroppedTargetsList(t *testing.T) {
 				},
 			},
 		}
-		sp, _                  = newScrapePool(cfg, app, 0, nil, false)
+		sp, _                  = newScrapePool(cfg, app, 0, nil, false, false, nil)
 		expectedLabelSetString = "{__address__=\"127.0.0.1:9090\", __scrape_interval__=\"0s\", __scrape_timeout__=\"0s\", job=\"dropMe\"}"
 		expectedLength         = 1
 	)
@@ -455,7 +454,7 @@ func TestScrapePoolTargetLimit(t *testing.T) {
 func TestScrapePoolAppender(t *testing.T) {
 	cfg := &config.ScrapeConfig{}
 	app := &nopAppendable{}
-	sp, _ := newScrapePool(cfg, app, 0, nil, false)
+	sp, _ := newScrapePool(cfg, app, 0, nil, false, false, nil)
 
 	loop := sp.newLoop(scrapeLoopOptions{
 		target: &Target{},
@@ -497,7 +496,7 @@ func TestScrapePoolRaces(t *testing.T) {
 	newConfig := func() *config.ScrapeConfig {
 		return &config.ScrapeConfig{ScrapeInterval: interval, ScrapeTimeout: timeout}
 	}
-	sp, _ := newScrapePool(newConfig(), &nopAppendable{}, 0, nil, false)
+	sp, _ := newScrapePool(newConfig(), &nopAppendable{}, 0, nil, false, false, nil)
 	tgts := []*targetgroup.Group{
 		{
 			Targets: []model.LabelSet{
@@ -591,6 +590,9 @@ func TestScrapeLoopStopBeforeRun(t *testing.T) {
 		1,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	// The scrape pool synchronizes on stopping scrape loops. However, new scrape
@@ -660,6 +662,9 @@ func TestScrapeLoopStop(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	// Terminate loop after 2 scrapes.
@@ -732,6 +737,9 @@ func TestScrapeLoopRun(t *testing.T) {
 		time.Second,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	// The loop must terminate during the initial offset if the context
@@ -784,6 +792,9 @@ func TestScrapeLoopRun(t *testing.T) {
 		time.Second,
 		100*time.Millisecond,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	go func() {
@@ -840,6 +851,9 @@ func TestScrapeLoopForcedErr(t *testing.T) {
 		time.Second,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	forcedErr := fmt.Errorf("forced err")
@@ -895,6 +909,9 @@ func TestScrapeLoopMetadata(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	defer cancel()
 
@@ -949,6 +966,9 @@ func simpleTestScrapeLoop(t testing.TB) (context.Context, *scrapeLoop) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	t.Cleanup(func() { cancel() })
 
@@ -1039,6 +1059,9 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	// Succeed once, several failures, then stop.
 	numScrapes := 0
@@ -1098,6 +1121,9 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	// Succeed once, several failures, then stop.
@@ -1161,6 +1187,9 @@ func TestScrapeLoopCache(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	numScrapes := 0
@@ -1240,6 +1269,9 @@ func TestScrapeLoopCacheMemoryExhaustionProtection(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	numScrapes := 0
@@ -1351,6 +1383,9 @@ func TestScrapeLoopAppend(t *testing.T) {
 			0,
 			0,
 			false,
+			nil,
+			nil,
+			false,
 		)
 
 		now := time.Now()
@@ -1438,7 +1473,7 @@ func TestScrapeLoopAppendForConflictingPrefixedLabels(t *testing.T) {
 					return mutateSampleLabels(l, &Target{labels: labels.FromStrings(tc.targetLabels...)}, false, nil)
 				},
 				nil,
-				func(ctx context.Context) storage.Appender { return app }, nil, 0, true, 0, nil, 0, 0, false,
+				func(ctx context.Context) storage.Appender { return app }, nil, 0, true, 0, nil, 0, 0, false, nil, nil, false,
 			)
 			slApp := sl.appender(context.Background())
 			_, _, _, err := sl.append(slApp, []byte(tc.exposedLabels), "", time.Date(2000, 1, 1, 1, 0, 0, 0, time.UTC))
@@ -1474,6 +1509,9 @@ func TestScrapeLoopAppendCacheEntryButErrNotFound(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	fakeRef := storage.SeriesRef(1)
@@ -1529,6 +1567,9 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	// Get the value of the Counter before performing the append.
@@ -1603,6 +1644,9 @@ func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -1648,6 +1692,9 @@ func TestScrapeLoopAppendStaleness(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -1696,6 +1743,9 @@ func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -1804,6 +1854,9 @@ metric_total{n="2"} 2 # {t="2"} 2.0 20000
 				0,
 				0,
 				false,
+				nil,
+				nil,
+				false,
 			)
 
 			now := time.Now()
@@ -1866,6 +1919,9 @@ func TestScrapeLoopAppendExemplarSeries(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -1915,6 +1971,9 @@ func TestScrapeLoopRunReportsTargetDownOnScrapeError(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
@@ -1948,6 +2007,9 @@ func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	scraper.scrapeFunc = func(ctx context.Context, w io.Writer) error {
@@ -1994,6 +2056,9 @@ func TestScrapeLoopAppendGracefullyIfAmendOrOutOfOrderOrOutOfBounds(t *testing.T
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Unix(1, 0)
@@ -2036,6 +2101,9 @@ func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now().Add(20 * time.Minute)
@@ -2129,7 +2197,7 @@ func TestTargetScrapeScrapeCancel(t *testing.T) {
 	}()
 
 	go func() {
-		_, err := ts.scrape(ctx, ioutil.Discard)
+		_, err := ts.scrape(ctx, io.Discard)
 		if err == nil {
 			errc <- errors.New("Expected error but got nil")
 		} else if ctx.Err() != context.Canceled {
@@ -2173,7 +2241,7 @@ func TestTargetScrapeScrapeNotFound(t *testing.T) {
 		client: http.DefaultClient,
 	}
 
-	_, err = ts.scrape(context.Background(), ioutil.Discard)
+	_, err = ts.scrape(context.Background(), io.Discard)
 	require.Contains(t, err.Error(), "404", "Expected \"404 NotFound\" error but got: %s", err)
 }
 
@@ -2290,6 +2358,9 @@ func TestScrapeLoop_RespectTimestamps(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -2328,6 +2399,9 @@ func TestScrapeLoop_DiscardTimestamps(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	now := time.Now()
@@ -2365,6 +2439,9 @@ func TestScrapeLoopDiscardDuplicateLabels(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	defer cancel()
 
@@ -2420,6 +2497,9 @@ func TestScrapeLoopDiscardUnnamedMetrics(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	defer cancel()
 
@@ -2512,7 +2592,7 @@ func TestReuseScrapeCache(t *testing.T) {
 			ScrapeInterval: model.Duration(5 * time.Second),
 			MetricsPath:    "/metrics",
 		}
-		sp, _ = newScrapePool(cfg, app, 0, nil, false)
+		sp, _ = newScrapePool(cfg, app, 0, nil, false, false, nil)
 		t1    = &Target{
 			discoveredLabels: labels.Labels{
 				labels.Label{
@@ -2693,6 +2773,9 @@ func TestScrapeAddFast(t *testing.T) {
 		0,
 		0,
 		false,
+		nil,
+		nil,
+		false,
 	)
 	defer cancel()
 
@@ -2722,7 +2805,7 @@ func TestReuseCacheRace(t *testing.T) {
 			ScrapeInterval: model.Duration(5 * time.Second),
 			MetricsPath:    "/metrics",
 		}
-		sp, _ = newScrapePool(cfg, app, 0, nil, false)
+		sp, _ = newScrapePool(cfg, app, 0, nil, false, false, nil)
 		t1    = &Target{
 			discoveredLabels: labels.Labels{
 				labels.Label{
@@ -2781,6 +2864,9 @@ func TestScrapeReportSingleAppender(t *testing.T) {
 		10*time.Millisecond,
 		time.Hour,
 		false,
+		nil,
+		nil,
+		false,
 	)
 
 	numScrapes := 0
@@ -2851,7 +2937,7 @@ func TestScrapeReportLimit(t *testing.T) {
 	}))
 	defer ts.Close()
 
-	sp, err := newScrapePool(cfg, s, 0, nil, false)
+	sp, err := newScrapePool(cfg, s, 0, nil, false, false, nil)
 	require.NoError(t, err)
 	defer sp.stop()
 
@@ -2980,6 +3066,9 @@ func TestScrapeLoopLabelLimit(t *testing.T) {
 			0,
 			0,
 			false,
+			nil,
+			nil,
+			false,
 		)
 
 		slApp := sl.appender(context.Background())
@@ -3018,7 +3107,7 @@ func TestTargetScrapeIntervalAndTimeoutRelabel(t *testing.T) {
 			},
 		},
 	}
-	sp, _ := newScrapePool(config, &nopAppendable{}, 0, nil, false)
+	sp, _ := newScrapePool(config, &nopAppendable{}, 0, nil, false, false, nil)
 	tgts := []*targetgroup.Group{
 		{
 			Targets: []model.LabelSet{{model.AddressLabel: "127.0.0.1:9090"}},
